// decoder.cc
//
// uda - universal disassembler
// Date: 2003-05-04
//
// Copyright (c) 2002-2003 Ales Smrcka 
//
// This program is licensed under GNU Library GPL. See the file COPYING.
//
// This module contains implementation of class Decoder (facade) used to
// decoding a section.
//

#include <string>
#include <ctype.h> //isprint()
#include "decoder.h"
#include "symtable.h"
#include "sections.h"
using namespace std;

const char *StartingTable="StartTable";
Decoder decoder;

DecoderInfo decinfo;

//! Default parameters for decoder
DecoderInfo::DecoderInfo()
{
    section=NULL;
    binfile=NULL;
    start=0;
    stop=0;
    size=0;
    print=true;
    address=true;
    hexdump=true;
    symbolic=true;
    onlyname=false;
    scanaddress=true;
}

DecoderInfo &DecoderInfo::assign(const DecoderInfo &decinfo)
{
    section=decinfo.section;
    binfile=decinfo.binfile;
    start=decinfo.start;
    stop=decinfo.stop;
    size=decinfo.size;
    print=decinfo.print;
    address=decinfo.address;
    hexdump=decinfo.hexdump;
    symbolic=decinfo.symbolic;
    onlyname=decinfo.onlyname;
    scanaddress=decinfo.scanaddress;
    return *this;
}

// Decode the section as hexadecimal output
void Decoder::HexOut()
{
    Address addr;
    Address stop;
    Section &section = *decinfo.section;
    if (decinfo.start==0)
        addr=section->addr;
    else
        addr=decinfo.start;
    if (decinfo.stop==0)
        if (decinfo.size==0)
            stop=addr+section->size;
        else
            stop=addr+decinfo.size;
    unsigned i;
    while (section.ValidAddress(addr) && addr<stop)
    {
        //print address
        cout<<""<<hex<<setw(8)<<setfill('0')<<addr<<": ";
        //print hex content
        for (i=0;i<16;i++)
            if (section.ValidAddress(addr+i) && addr+i<stop)
            {
                if (i%4==0)
                    cout<<" ";
                cout<<hex<<setw(2)<<setfill('0')<<unsigned(section[addr+i])<<" ";
            }
            else
                break;
        for (;i<16;i++) //fill with " "
        {
            if (i%4==0) cout<<" ";
            cout<<"   ";
        }
        //print ASCII content
        cout<<" ";
        for (i=0;i<16;i++)
            if (section.ValidAddress(addr+i) && addr+i<stop)
            {
                if (isprint(section[addr+i]))
                    cout<<section[addr+i];
                else
                    cout<<'.';
            }
            else
                break;
        cout<<endl;
        addr+=16;
    }    
}

//prints a line with symbolic instruction
void Decoder::print_line(Address addr, unsigned size, string &symbolic_instruction)
{
    unsigned labellength=0;
    unsigned space=16;
    Section &section = *decinfo.section;
    if (symbols->defined(addr))
    {
        string s=(*symbols)[addr];
        if (s.substr(0, 5)!="label")
            cout<<endl<<s<<":"<<endl;
        else
        {
            cout<<s<<":";
            if (!decinfo.onlyname)
                cout<<endl;
            labellength=s.length()+1;
        }
    }
    if (decinfo.address)
        cout<<"0x"<<setw(8)<<setfill('0')<<hex<<addr<<":";
    if (decinfo.hexdump)
    {
        unsigned i;
        cout<<"  ";
        for (i=0;i<size;i++)
            cout<<hex<<setw(2)<<setfill('0')<<unsigned(section[addr+i])<<" ";
        for (;i<8;i++)
            cout<<"   ";
    }
    if (decinfo.hexdump || decinfo.address)
        space=10;
    if (decinfo.onlyname)
        for (unsigned i=0;i<space-labellength;i++)
            cout<<" ";
    if (decinfo.symbolic)
        cout<<symbolic_instruction<<endl;
}

// Decode the section
void Decoder::Decode()
{
    Address addr;
    Address stop;
    if (decinfo.section==NULL)
        throw Exception("Decoder:section isn't set\n");
    Section &section = *decinfo.section;
    if (!section->executable)
        if (decinfo.print)
            HexOut();
        else;
    else
    {
        if (decinfo.start==0)
            addr=section->addr;
        else
            addr=decinfo.start;
        if (decinfo.stop==0)
            if (decinfo.size==0)
                stop=addr+section->size;
            else
                stop=addr+decinfo.size;
        else
            stop=decinfo.stop;
        unsigned ins_size;
        string symbolic_instruction;
        Tables &tables=*instructionset.tables;
        Variables &variables=*instructionset.variables;
        string starttable;
        AbstractTable *table;
        while (section.ValidAddress(addr) && addr<stop)
        {
            table=tables[variables[StartingTable]];
            symbolic_instruction=table->Decode(section, addr, ins_size);
            if (decinfo.print)
                print_line(addr, ins_size, symbolic_instruction);
            addr+=ins_size;
            tables.Reset();
            variables.Reset();
        }
    }
}
