// instrset.h
//
// uda - universal disassembler
// Date: 2003-05-04
//
// Copyright (c) 2002-2003 Ales Smrcka 
//
// This program is licensed under GNU Library GPL. See the file COPYING.
//
// An interface of loading instruction set from file and creating tables and
// variables. This file including builtin tables in namespace BuiltinTables.
// Tables, Table, TableItem include core of decoding.

#ifndef instrsetH
#define instrsetH

#include <iomanip>
#include <iostream>
#include <fstream>
#include <map>
#include <string>
#include "shared.h"
#include "udaclasses.h"
#include "sections.h"
using namespace std;

class InstrSetException;
class  ScannerException;
class  ParserException;
class  InstructionException;

class InstructionSet;
// Scanner
class Scanner;
// Parser ist. map
class Parser;
// Tables of instructions map. Map of Table
class Tables;
// Abstract table, parent of BuildinTable and Table
class AbstractTable;
// Builtin table
class  BuiltinTable;
// One table of Tables including TableItem
class  Table;
// Item of Table
class TableItem;
// Map of Variable
class Variables;
// Item of Variables
class Variable;

extern const string instrsetdirectory;
extern const string defaultinstrset;
//! Instruction set - singleton
extern InstructionSet instructionset;

class InstrSetException : public Exception {
protected:
    int line;
public:
    InstrSetException(string s, int l=0) : Exception(s) { line=l; }
    virtual ~InstrSetException() {}
    virtual void Print() { cerr<<"Instruction set error at line "<<line<<" : "<<value<<endl; }
};

class ScannerException : public InstrSetException {
public:
    ScannerException(string s, int l) : InstrSetException(s, l) {}
    virtual ~ScannerException() {}
    void Print() { cerr<<"Scanner error at line "<<line<<" : "<<value<<endl; }
};

class ParserException : public InstrSetException {
public:
    ParserException(string s, int l) : InstrSetException(s, l) {}
    virtual ~ParserException() {}
    void Print() { cerr<<"Parser error at line "<<line<<" : "<<value<<endl; }
};

class InstructionException : public InstrSetException {
public:
    string text;
    InstructionException(string s, string t) : InstrSetException(s) { text=t; }
    virtual ~InstructionException() {}
    void Print() { cerr<<"Instruction decoder error: "<<value<<": "<<text<<endl; }
};

//! InstructionSet. (Facade, Singleton)
class InstructionSet {
    InstructionSet(const InstructionSet&) {}
public:
    Tables *tables;
    Variables *variables;
    //! creates tables and variables
    InstructionSet();
    //! frees tables and variables
    ~InstructionSet();
    //! fills tables and variables
    int LoadFile(const string filename);
    //! load instruction set from file corresponding to processor
    void Load(const string processor);
    //! clear instruction set
    void Clear();
};

//! tk_type = Token type
enum tk_type {
    //! end of file
    tk_end=0,
    //! token is a number
    tk_number,
    //! token is a string literal
    tk_text,
    //! token is an identificator
    tk_ident,
    //! token is '=' character
    tk_equal,
    //! token is '+' character
    tk_plus,
    //! token is ',' character
    tk_comma,
    //! token is ';' character
    tk_semicol,
    //! token is '{' character
    tk_parop,
    //! token is '}' character
    tk_parcl
};

//! Scans input for tokens
class Scanner {
    Scanner();
    int unget; //unget char
    int _line;
protected:
    const char *_filename;
    ifstream input; //input stream from file with instruction set
    int getc();
    void ungetc(int c);
public:
    //! Token type returned by Scanner
    class Token {
    public:
        //! type of this token
        tk_type type;
        //! string value
        string str_value;
        //! unsigned integer value
        unsigned int_value;
        Token(tk_type t=tk_end, unsigned i=0) { type=t; int_value=i; }
        operator tk_type() { return type; }
    };
    //! Returns actual row in the input stream
    int line() { return _line; }
    //! Creates scanner
    Scanner(const char *filename);
    //! Reads next token. tk_none type is returned when no token is in input stream.
    Token Scan();
};

//! Creates tables and variables
class Parser {
    Scanner &scanner;
    Tables &tables;
    Variables &variables;
    Parser();
    void Error();
public:
    //! Creates parser connected to scanner sc, tables t and variables v
    Parser(Scanner &sc, Tables &t, Variables &v) : scanner(sc), tables(t), variables(v) {}
    //! Runs parser to create new tables and variables
    void Parse();
};

typedef map<string, AbstractTable*> MTables;

//! Map of Table indexed by its name. (Builder)
class Tables {
    MTables _tables;
public:
    Variables *variables;
    Tables(Variables *v) { variables=v; }
    ~Tables();
    //! Add(name) creates new table (Table) and returns it.
    Table *Add(string name);
    //! Add(name, table) - useful for builtin tables
    AbstractTable *&Add(string name, AbstractTable *table);
    //! Returns appropriate table. Throws exception when table does not exist.
    AbstractTable *&operator [](string index);
    //! Resets all tables to their default values (because of lengths of the items)
    void Reset();
    //! Clears all tables from memory
    void Clear();
};

//! AbstractTable - parent for BuiltinTable and Table
class AbstractTable {
public:
    virtual ~AbstractTable() {}
    //! Decode actual bytes. input: data and address, output: decoded string and size of decoded data
    virtual string Decode(Section &data, Address addr, unsigned &size);
    //! Resets all items' length
    virtual void Reset() {}
};

//! Parent of all builtin tables
class BuiltinTable : public AbstractTable {};

//! Table with 256 items (TableItem). Item is indexed by a byte in data section.
class Table : public AbstractTable {
    TableItem *_items[256];
public:
    Tables *tables;
    Variables *variables;
    Table(Tables *t=NULL, Variables *v=NULL); //for all items: item=NULL
    virtual ~Table(); //free all items
    TableItem *Add(unsigned mask, unsigned tag, string value, string length);
    //! Decode actual byte. input: data and address, output: decoded string and size of decoded data
    string Decode(Section &data, Address addr, unsigned &size);
//    TableItem &operator [](unsigned index) { return *_items[index]; }
    //! Resets all items' length
    void Reset();
};

//! An item of Table consisting of value and length array of decoded bytes
class TableItem {
    void Error(const char *c=NULL);
    //! Returns sum of values of all valid lengths
    int CountLength(unsigned lens[]);
    //! Returns sum of values of lengths len (e.g. "1+b" returns 1+lens[2]; "" returns 0)
    int CountLength(unsigned lens[], string len);
    // Decodes the text (c) with others tables and variables. Fills the length array.
    // lindex is actual index to length array. c is the text. is terminated character.
    // Returns substituted string.
    string Norm(Section &data, Address addr, unsigned lens[], unsigned &lindex, const char *&c, char t=0);
public:
    Tables &tables;
    Variables &variables;
    string value;
    struct Lengths {
        unsigned first_value;
        bool valid[26];
    } lengths;
    TableItem(Tables &t, Variables &v, string value, string lengthstr);
    //! Decode actual byte. input: data and address, output: decoded string and size of decoded data
    string Decode(Section &data, Address addr, unsigned &size);
};

typedef map<string, Variable> MVariables;

//! Map of Variable indexed by its name. (Builder)
class Variables {
    MVariables _variables;
public:
    //! Creates new variable
    Variable &Add(string name, string defaultvalue="");
    //! Returns appropriate variable, throws an exception if variable does not exist.
    Variable &operator [](string index);
    //! Resets all variables to their default value
    void Reset();
    //! Clears all variables from memory
    void Clear();
};

//! Variable is value and default value for one variable
class Variable {
public:
    string value;
    string defaultvalue;
    Variable(string defaultvalue="") { Variable::value=defaultvalue; Variable::defaultvalue=defaultvalue; }
    void Reset() { value=defaultvalue; }
    operator string() { return value; }
};

namespace BuiltinTables {

//////////////////////////////////////////////////////////////////////////////////////
//
// Little Endian - Immediate data
//
//////////////////////////////////////////////////////////////////////////////////////

//! Table - Little Endian - immediate byte
class Ib : public BuiltinTable {
public:
    virtual string Decode(Section &data, Address addr, unsigned &size);
};
//! Table - Little Endian - immediate word
class Iw : public BuiltinTable {
public:
    virtual string Decode(Section &data, Address addr, unsigned &size);
};
//! Table - Little Endian - immediate dword
class Id : public BuiltinTable {
public:
    virtual string Decode(Section &data, Address addr, unsigned &size);
};
//! Table - Little Endian - immediate byte as word
class Ibw : public BuiltinTable {
public:
    virtual string Decode(Section &data, Address addr, unsigned &size);
};
//! Table - Little Endian - immediate byte as dword
class Ibd : public BuiltinTable {
public:
    virtual string Decode(Section &data, Address addr, unsigned &size);
};

//////////////////////////////////////////////////////////////////////////////////////
//
// Little Endian - Signed Immediate daat
//
//////////////////////////////////////////////////////////////////////////////////////

//! Table - Signed Little Endian - immediate byte
class SIb : public BuiltinTable {
public:
    virtual string Decode(Section &data, Address addr, unsigned &size);
};
//! Table - Signed Little Endian - immediate word
class SIw : public BuiltinTable {
public:
    virtual string Decode(Section &data, Address addr, unsigned &size);
};
//! Table - Signed Little Endian - immediate dword
class SId : public BuiltinTable {
public:
    virtual string Decode(Section &data, Address addr, unsigned &size);
};
//! Table - Signed Little Endian - immediate byte as word
class SIbw : public BuiltinTable {
public:
    virtual string Decode(Section &data, Address addr, unsigned &size);
};
//! Table - Signed Little Endian - immediate byte as dword
class SIbd : public BuiltinTable {
public:
    virtual string Decode(Section &data, Address addr, unsigned &size);
};

//////////////////////////////////////////////////////////////////////////////////////
//
// Little Endian - Relative address
//
//////////////////////////////////////////////////////////////////////////////////////

//! Table - Little Endian - relative address - byte
class Jb : public BuiltinTable {
public:
    virtual string Decode(Section &data, Address addr, unsigned &size);
};
//! Table - Little Endian - relative address - word
class Jw : public BuiltinTable {
public:
    virtual string Decode(Section &data, Address addr, unsigned &size);
};
//! Table - Little Endian - relative address - dword
class Jd : public BuiltinTable {
public:
    virtual string Decode(Section &data, Address addr, unsigned &size);
};

//////////////////////////////////////////////////////////////////////////////////////
//
// Little Endian - Absolute address
//
//////////////////////////////////////////////////////////////////////////////////////

//! Table - Little Endian - absolute segment address - word:word
class Apw : public BuiltinTable {
public:
    virtual string Decode(Section &data, Address addr, unsigned &size);
};
//! Table - Little Endian - absolute segment address - word:dword
class Apd : public BuiltinTable {
public:
    virtual string Decode(Section &data, Address addr, unsigned &size);
};
//! Table - Little Endian - absolute address - word
class Aw : public BuiltinTable {
public:
    virtual string Decode(Section &data, Address addr, unsigned &size);
};
//! Table - Little Endian - absolute address - dword
class Ad : public BuiltinTable {
public:
    virtual string Decode(Section &data, Address addr, unsigned &size);
};

//////////////////////////////////////////////////////////////////////////////////////
//
// Big Endian - Immediate data
//
//////////////////////////////////////////////////////////////////////////////////////

//! Table - Big Endian - immediate byte
class BIb : public BuiltinTable {
public:
    virtual string Decode(Section &data, Address addr, unsigned &size);
};
//! Table - Big Endian - immediate word
class BIw : public BuiltinTable {
public:
    virtual string Decode(Section &data, Address addr, unsigned &size);
};
//! Table - Big Endian - immediate dword
class BId : public BuiltinTable {
public:
    virtual string Decode(Section &data, Address addr, unsigned &size);
};
//! Table - Big Endian - immediate byte as word
class BIbw : public BuiltinTable {
public:
    virtual string Decode(Section &data, Address addr, unsigned &size);
};
//! Table - Big Endian - immediate byte as dword
class BIbd : public BuiltinTable {
public:
    virtual string Decode(Section &data, Address addr, unsigned &size);
};

//////////////////////////////////////////////////////////////////////////////////////
//
// Big Endian - Signed Immediate daat
//
//////////////////////////////////////////////////////////////////////////////////////

//! Table - Signed Big Endian - immediate byte
class SBIb : public BuiltinTable {
public:
    virtual string Decode(Section &data, Address addr, unsigned &size);
};
//! Table - Signed Big Endian - immediate word
class SBIw : public BuiltinTable {
public:
    virtual string Decode(Section &data, Address addr, unsigned &size);
};
//! Table - Signed Big Endian - immediate dword
class SBId : public BuiltinTable {
public:
    virtual string Decode(Section &data, Address addr, unsigned &size);
};
//! Table - Signed Big Endian - immediate byte as word
class SBIbw : public BuiltinTable {
public:
    virtual string Decode(Section &data, Address addr, unsigned &size);
};
//! Table - Signed Big Endian - immediate byte as dword
class SBIbd : public BuiltinTable {
public:
    virtual string Decode(Section &data, Address addr, unsigned &size);
};

//////////////////////////////////////////////////////////////////////////////////////
//
// Big Endian - Relative address
//
//////////////////////////////////////////////////////////////////////////////////////

//! Table - Big Endian - relative address - byte
class BJb : public BuiltinTable {
public:
    virtual string Decode(Section &data, Address addr, unsigned &size);
};
//! Table - Big Endian - relative address - word
class BJw : public BuiltinTable {
public:
    virtual string Decode(Section &data, Address addr, unsigned &size);
};
//! Table - Big Endian - relative address - dword
class BJd : public BuiltinTable {
public:
    virtual string Decode(Section &data, Address addr, unsigned &size);
};

//////////////////////////////////////////////////////////////////////////////////////
//
// Big Endian - Absolute address
//
//////////////////////////////////////////////////////////////////////////////////////

//! Table - Big Endian - absolute segment address - word:word
class BApw : public BuiltinTable {
public:
    virtual string Decode(Section &data, Address addr, unsigned &size);
};
//! Table - Big Endian - absolute segment address - word:dword
class BApd : public BuiltinTable {
public:
    virtual string Decode(Section &data, Address addr, unsigned &size);
};
//! Table - Big Endian - absolute address - word
class BAw : public BuiltinTable {
public:
    virtual string Decode(Section &data, Address addr, unsigned &size);
};
//! Table - Big Endian - absolute address - dword
class BAd : public BuiltinTable {
public:
    virtual string Decode(Section &data, Address addr, unsigned &size);
};

} //namespace BuiltinTables
#endif
