// plugins.cc
//
// uda - universal disassembler
// Date: 2003-05-04
//
// Copyright (c) 2002-2003 Ales Smrcka 
//
// This program is licensed under GNU Library GPL. See the file COPYING.
//
// This module contains administration of plugins (loading, syncing).
//

#include <iostream>
#include <dlfcn.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <dirent.h>
#include <libgen.h>
#include "plugins.h"
#include "shared.h"
#include "udaclasses.h"
using namespace std;

const char *pluginsdir = "plugins/";
const char *configfile = "plugins/list";

// plugins array (not NULL are relevant)
VPlugins plugins;

int EvEmpty_data(void*)
{
    return 0;
}

int EvEmpty_void()
{
    return 0;
}

//////////////////////////////////////////////////////////////////////
// Class Plugin

Plugin::Plugin()
{
    dynfile=NULL;
    IsFileDefined=(Ev_IsFileDefined)&EvEmpty_data;
    LoadSections=(Ev_LoadSections)&EvEmpty_data;
    LoadSymbols=(Ev_LoadSymbols)&EvEmpty_void;
}

Plugin::~Plugin()
{
    Close();
}

//! Load plugin information. Returns 1 on success.
int Plugin::Open(const char *file)
{
    dynfile = dlopen(file, RTLD_LAZY);
    if (dynfile==NULL)
        return 0;
    Splugin_init plugin_init = (Splugin_init)dlsym(dynfile, "plugin_init");
    const char *error;
    if ((error=dlerror())!=NULL)
    {
        debug("plugin_init function mistakes in "<<basename((char*)file)<<endl);
        dlclose(dynfile);
        return 0;
    }
    plugin_init(type, ident, syntax, Params);
    if (type&PLUGIN_SECTIONS)
    {
        IsFileDefined = (Ev_IsFileDefined)dlsym(dynfile, "is_file_defined");
        if ((error=dlerror())!=NULL)
        {
            debug("is_file_defined function mistakes in "<<basename((char*)file)<<endl);
            dlclose(dynfile);
            return 0;
        }
        LoadSections = (Ev_LoadSections)dlsym(dynfile, "load_sections");
        if ((error=dlerror())!=NULL)
        {
            debug("load_sections function mistakes in "<<basename((char*)file)<<endl);
            dlclose(dynfile);
            return 0;
        }
    }
    if (type&PLUGIN_SYMTABLE)
    {
        LoadSymbols = (Ev_LoadSymbols)dlsym(dynfile, "load_symbols");
        if ((error=dlerror())!=NULL)
        {
            debug("load_symbols function mistakes in "<<basename((char*)file)<<endl);
            dlclose(dynfile);
            return 0;
        }
    }
    filename=basename((char*)file);
    debug("Loaded plugin: "<<filename<<endl);
    return 1;
}

void Plugin::Close()
{
    if (dynfile!=NULL)
        dlclose(dynfile);
}

// Print plugin information
void Plugin::Info()
{
    cout<<"Plugin "<<filename<<":"<<endl;
    cout<<"  ID="<<ident<<endl;
    cout<<"  attributes=";
    if (type&PLUGIN_SECTIONS)
        cout<<"PLUGIN_SECTIONS";
    if (type&~PLUGIN_SECTIONS)
        cout<<", ";
    if (type&PLUGIN_SYMTABLE)
        cout<<"PLUGIN_SYMTABLE";
    cout<<endl;
    cout<<"  options=";
    if (syntax!=NULL)
        cout<<syntax;
    else
        cout<<"<none>";
    cout<<endl;
}

//////////////////////////////////////////////////////////////////////
// OpenPlugins() and ClosePlugins()

// Load all plugins
int OpenPlugins()
{
    int actualpl=0;
    FILE *cfg;
    char file[1024];
    Plugin *pl;
    debug("Opening plugins from config file...\n");
    //open config file (list of plugins)
    if ((cfg=fopen((basedir+configfile).c_str(), "r"))!=NULL)
    {
        while (fgets(file, 1023, cfg)!=NULL)
        {
            //get C string from line
            file[1023]=0;
            for (int i=0;i<1024;i++)
                if (file[i]=='\n')
                {
                    file[i]=0;
                    break;
                }
            //create plugin info
            pl=new Plugin;
            //load plugin info
            debug("Plugins::Open(): try to open "<<file<<endl);
            if (!pl->Open(file) && !pl->Open((basedir+pluginsdir+file).c_str()))
            {
                debug("Can not open plugin "<<file<<endl);
                delete pl; //bad plugin
            }
            else
                plugins.push_back(pl);
        } //while load line
        fclose(cfg);
    }
    debug("Opening plugins from plugins directory...\n");
    
    //load plugins from plugin directory
    DIR *dir;
    struct dirent *dp;
    struct stat buf;
    char olddir[1024];
    getcwd(olddir, 1023);
    if (chdir((basedir+pluginsdir).c_str()))
        return actualpl;
    if ((dir=opendir("."))==NULL)
        return actualpl;
    file[0]='.';file[1]='/';
    //read files
    while ((dp=readdir(dir))!=NULL)
        if ((stat(dp->d_name, &buf)==0) && (S_ISREG(buf.st_mode)))//is regular (symlink included)
        {
            register bool already_loaded=false;
            VPlugins::iterator i;
            for (i=plugins.begin();i!=plugins.end();i++)
                if ((already_loaded = (*i)->filename==dp->d_name))
                    break;
            if (!already_loaded)
            {
                strncpy(&file[2], dp->d_name, 1020);
                //create plugin info
                pl=new Plugin;
                //load plugin info
                debug("Plugins::Open(): try to open "<<&file[2]<<endl);
                if (!pl->Open(file))
                {
                    debug("Can not open plugin "<<file<<endl);
                    delete pl; //bad plugin
                }
                else
                    plugins.push_back(pl);
            }
        }
    closedir(dir);
    chdir(olddir);
    return plugins.size();
} //OpenPlugins

void ClosePlugins()
{
    VPlugins::iterator i;
    for (i=plugins.begin();i!=plugins.end();i++)
        delete *i;
    plugins.clear();
}

void PluginsInfo()
{
    VPlugins::iterator i;
    for (i=plugins.begin();i!=plugins.end();i++)
        (*i)->Info();
}

